"""Run tests on documentation tests(doctest).

Currently setup this way b/c pytest does not support "load_tests" protocol.
"""
import unittest

import doctest

import mochart.utils


def load_tests(loader, tests, ignore):
    """Load doctests as unit test suite."""
    tests.addTests(doctest.DocTestSuite(mochart.utils))
    return tests


if __name__ == "__main__":
    unittest.main()



import os

import pytest


def get_html_fixture(provider):
    filename = f"{provider}.html"
    with open(os.path.join(os.path.dirname(__file__), "fixtures", filename), "r") as f:
        return f.read()


@pytest.fixture
def melon_realtime():
    return get_html_fixture("melon_realtime")


@pytest.fixture
def mnet_realtime():
    return get_html_fixture("mnet_realtime")


@pytest.fixture
def gaon_week():
    return get_html_fixture("gaon_week")


@pytest.fixture
def naver_day():
    return get_html_fixture("naver_day")


@pytest.fixture
def oricon_day():
    return get_html_fixture("oricon_day")



from mochart import utils


BASE_URL = "https://music.naver.com/listen/top100.nhn?domain=TOTAL"
SELECTOR = "tr"
TIMEZONE = "Asia/Seoul"


def parser(rows):
    """Parse texts accordingly from Naver table."""
    ranks = []
    for row in rows[2:]:
        rank = {}
        rank["title"] = utils.group_multiples(
            row.select("td.name > a._title > span"))
        rank["artist"] = utils.group_multiples(row.select("td._artist > a"))
        ranks.append(rank)
    return ranks


def realtime():
    """Return realtime rankings."""
    url = f"{BASE_URL}&duration=1h"
    return utils.get_ranks(url, SELECTOR, parser)


def day():
    """Return rankings for the day."""
    url = f"{BASE_URL}&duration=1d"
    return utils.get_ranks(url, SELECTOR, parser)


def week(day_time=None):
    """Return rankings for given week."""
    url = f"{BASE_URL}&duration=7d"

    if day_time is not None:
        base_url = "https://music.naver.com/listen/history/index.nhn?type=TOTAL"

        local_dt = utils.localize_time(day_time, TIMEZONE)
        year = local_dt.strftime("%Y")
        month = local_dt.strftime("%m")
        # 0 == all, so the first week is 1.
        week = utils.get_weeks_in(local_dt)

        url = f"{base_url}&year={year}&month={month}&week={week}"

    return utils.get_ranks(url, SELECTOR, parser)


def month(day_time=None):
    """Return rankings for given month."""
    url = f"{BASE_URL}&duration=28d"
    if day_time is not None:
        base_url = "https://music.naver.com/listen/history/index.nhn?type=TOTAL"

        local_dt = utils.localize_time(day_time, TIMEZONE)
        year = local_dt.strftime("%Y")
        month = local_dt.strftime("%m")
        # 0 == all, so the first week is 1.
        week = 0

        url = f"{base_url}&year={year}&month={month}&week={week}"

    return utils.get_ranks(url, SELECTOR, parser)



from mochart import utils


BASE_URL = "https://www.oricon.co.jp/rank/js"
SELECTOR = "section.box-rank-entry"
TIMEZONE = "Asia/Tokyo"


def parser(rows):
    """Parse texts accordingly from Oricon table."""
    ranks = []
    for row in rows:
        rank = {}
        rank["title"] = utils.group_multiples(row.select(".title"))
        rank["artist"] = utils.group_multiples(row.select(".name"))
        ranks.append(rank)
    return ranks


def day(day_time=None):
    base_url = f"{BASE_URL}/d"
    local_dt = utils.localize_time(day_time, TIMEZONE)
    url = utils.append_date_string(
        base_url, local_dt, date_format="%Y-%m-%d", trailing_slash=True)
    return utils.get_ranks(url, SELECTOR, parser)


def week(day_time=None):
    base_url = f"{BASE_URL}/w"
    local_dt = utils.localize_time(day_time, TIMEZONE)
    beg, _ = utils.get_week_dates(local_dt)
    url = utils.append_date_string(
        base_url, local_dt, date_format="%Y-%m-%d", trailing_slash=True)
    return utils.get_ranks(url, SELECTOR, parser)


def month(day_time=None):
    base_url = f"{BASE_URL}/m"
    local_dt = utils.localize_time(day_time, TIMEZONE)
    beg, _ = utils.get_week_dates(local_dt)
    url = utils.append_date_string(
        base_url, local_dt, date_format="%Y-%m", trailing_slash=True)
    return utils.get_ranks(url, SELECTOR, parser)


def year(day_time=None):
    base_url = f"{BASE_URL}/y"
    local_dt = utils.localize_time(day_time, TIMEZONE)
    beg, _ = utils.get_week_dates(local_dt)
    url = utils.append_date_string(
        base_url, local_dt, date_format="%Y", trailing_slash=True)
    return utils.get_ranks(url, SELECTOR, parser)



"""Unit tests for utility functions."""
from unittest.mock import MagicMock

import mochart.utils as utils


def test_get_html_document(monkeypatch):
    """Assert GET request is sent with URL and headers."""
    mock_requests = MagicMock()
    monkeypatch.setattr(utils, "requests", mock_requests)

    url = "some URL string"
    utils.get_html_document(url)

    assert url in mock_requests.get.call_args[0]
    assert "headers" in mock_requests.get.call_args[1]



"""Parse ranks from Melon Music Chart."""
from mochart import utils


BASE_URL = "https://www.melon.com/chart"
SELECTOR = "tr"


def parser(rows):
    """Parse texts accordingly from Melon table."""

    def remove_dups(rows):
        return rows[len(rows) // 2:]

    def parse(selector):
        return map(
            (
                lambda row:
                    utils.group_multiples(
                        remove_dups(
                            row.select(selector)))
            ),
            rows[1:]
        )

    return [{
        "title": t[0],
        "artist": t[1],
        "album": t[2]
    } for t in zip(
        parse("div.ellipsis.rank01 a"),
        parse("div.ellipsis.rank02 a"),
        parse("div.ellipsis.rank03 a"),
    )]


def realtime():
    """Get latest real-time ranks.

    NOTE: According to Melon's public announcement,
          "real-time" work will pause for 6 hours every day from 1AM - 7AM.
          Reference:
          https://www.melon.com/customer/announce/infomAnnounce.htm?seq=668
    """
    url = f"{BASE_URL}/index.htm"
    return utils.get_ranks(url, SELECTOR, parser)


def trend(day_time=None):
    """Get latest trending ranks.

    NOTE: Historical value refreshes daily.
    """
    base_url = f"{BASE_URL}/rise/index.htm"
    local_dt = utils.localize_time(day_time, "Asia/Seoul")
    url = utils.append_date_string(
        base_url, local_dt, date_key="dayTime", date_format="%Y%m%d%H")
    return utils.get_ranks(url, SELECTOR, parser)


def day():
    """Get latest daily ranks."""
    url = f"{BASE_URL}/day/index.htm"
    return utils.get_ranks(url, SELECTOR, parser)


def week():
    """Get latest weekly ranks."""
    url = f"{BASE_URL}/week/index.htm"
    return utils.get_ranks(url, SELECTOR, parser)


def month():
    """Get latest monthly ranks."""
    url = f"{BASE_URL}/month/index.htm"
    return utils.get_ranks(url, SELECTOR, parser)



from datetime import datetime, timedelta

from mochart import melon, mnet, naver, gaon, oricon


def print_first_five(ranks):
    ranks = list(ranks)
    for rank in ranks[:5]:
        print(rank)


provider = melon

# print(" - - - - - NOW")
# ranks = provider.peek()
# print_first_five(ranks)

print(" - - - - - NOW")
ranks = provider.realtime()
# print_first_five(ranks)
from pprint import pprint
pprint(ranks)

import json
with open("melon_realtime.json", "w", encoding="utf8") as f:
    f.write(json.dumps(ranks, ensure_ascii=False))

# print(" - - - - - 2 hours ago")
# four_hours_ago = datetime.now() - timedelta(hours=2)
# ranks = provider.trend(four_hours_ago)
# print_first_five(ranks)

# print(" - - - - - 4 hours ago")
# four_hours_ago = datetime.now() - timedelta(hours=4)
# ranks = provider.realtime(four_hours_ago)
# print_first_five(ranks)

# print(" - - - - - today")
# ranks = provider.day(datetime.now())
# print_first_five(ranks)


# print(" - - - - - weekly today")
# week_ago = datetime.now()
# ranks = provider.week()
# print_first_five(ranks)

# print(" - - - - - 1 week ago")
# week_ago = datetime.now() - timedelta(days=7)
# ranks = provider.week(week_ago)
# print_first_five(ranks)

# print(" - - - - - 2 week ago")
# two_weeks_ago = datetime.now() - timedelta(days=14)
# ranks = provider.week(two_weeks_ago)
# print_first_five(ranks)

# print(" - - - - - 1 month ago")
# month_ago = datetime.now() - timedelta(days=31)
# ranks = provider.month(month_ago)
# print_first_five(ranks)

# print(" - - - - - 5 month ago")
# month_ago = datetime.now() - timedelta(days=30*5)
# ranks = provider.month(month_ago)
# print_first_five(ranks)

# print(" - - - - - 1 year ago")
# year_ago = datetime.now() - timedelta(days=365)
# ranks = provider.year(year_ago)
# print_first_five(ranks)

# print(" - - - - - 5 year ago")
# year_ago = datetime.now() - timedelta(days=365*5)
# ranks = provider.year(year_ago)
# print_first_five(ranks)



from datetime import datetime, timedelta
from unittest.mock import MagicMock

import pytest

from mochart import (melon, mnet, naver, gaon, oricon)


PROVIDERS = [melon, mnet, naver, gaon, oricon]


@pytest.mark.integration
def test_sanity_checks():
    """Assert some ranks are fetched."""
    two_days_ago = datetime.now() - timedelta(days=2)
    for provider in PROVIDERS:
        try:
            assert len(provider.day(two_days_ago)) > 5
        except TypeError:
            assert len(provider.day()) > 5
        except AttributeError:
            assert len(provider.week(two_days_ago)) > 5


def test_gaon_parser(monkeypatch, gaon_week):
    """Assert parser extracts some ranks."""
    mock_get_html_document = MagicMock(return_value=gaon_week)
    monkeypatch.setattr(gaon.utils, "get_html_document", mock_get_html_document)
    ranks = gaon.week()
    assert len(ranks) > 0


def test_melon_parser(monkeypatch, melon_realtime):
    """Assert parser extracts some ranks."""
    mock_get_html_document = MagicMock(return_value=melon_realtime)
    monkeypatch.setattr(melon.utils, "get_html_document", mock_get_html_document)
    ranks = melon.realtime()
    assert len(ranks) > 0


def test_mnet_parser(monkeypatch, mnet_realtime):
    """Assert parser extracts some ranks."""
    mock_get_html_document = MagicMock(return_value=mnet_realtime)
    monkeypatch.setattr(mnet.utils, "get_html_document", mock_get_html_document)
    ranks = mnet.realtime()
    assert len(ranks) > 0


def test_naver_parser(monkeypatch, naver_day):
    """Assert parser extracts some ranks."""
    mock_get_html_document = MagicMock(return_value=naver_day)
    monkeypatch.setattr(naver.utils, "get_html_document", mock_get_html_document)
    ranks = naver.day()
    assert len(ranks) > 0


def test_oricon_parser(monkeypatch, oricon_day):
    """Assert parser extracts some ranks."""
    mock_get_html_document = MagicMock(return_value=oricon_day)
    monkeypatch.setattr(oricon.utils, "get_html_document", mock_get_html_document)
    ranks = oricon.day()
    assert len(ranks) > 0



from mochart import utils


BASE_URL = "http://www.mnet.com/chart/TOP100"
SELECTOR = "td.MMLItemTitle > div > div.MMLITitle_Box.info"
TIMEZONE = "Asia/Seoul"


def parser(rows):
    """Parse texts accordingly from Mnet table."""
    ranks = []
    for row in rows:
        rank = {}
        rank["title"] = utils.group_multiples(row.select("a.MMLI_Song"))
        rank["album"] = utils.group_multiples(row.select("a.MMLIInfo_Album"))
        # Requires a bit of tricks due to multiple artists.
        rank["artist"] = utils.group_multiples(row.select("a.MMLIInfo_Artist"))
        ranks.append(rank)
    return ranks


def realtime(day_time=None):
    local_dt = utils.localize_time(day_time, TIMEZONE)
    url = utils.append_date_string(BASE_URL, local_dt, date_format="%Y%m%d%H")
    return utils.get_ranks(
        url, SELECTOR, parser)


def day(day_time=None):
    local_dt = utils.localize_time(day_time, TIMEZONE)
    url = utils.append_date_string(BASE_URL, local_dt, date_format="%Y%m%d")
    return utils.get_ranks(
        url, SELECTOR, parser)


def week(day_time=None):
    local_dt = utils.localize_time(day_time, TIMEZONE)
    beg, end = utils.get_week_dates(local_dt)
    date_format = "%Y%m%d"
    url = f"{BASE_URL}/{beg.strftime(date_format)}-{end.strftime(date_format)}"
    return utils.get_ranks(
        url, SELECTOR, parser)


def month(day_time=None):
    local_dt = utils.localize_time(day_time, TIMEZONE)
    url = utils.append_date_string(BASE_URL, local_dt, date_format="%Y%m")
    return utils.get_ranks(
        url, SELECTOR, parser)


def year(day_time=None):
    local_dt = utils.localize_time(day_time, TIMEZONE)
    url = utils.append_date_string(BASE_URL, local_dt, date_format="%Y")
    return utils.get_ranks(
        url, SELECTOR, parser)



"""Utility functions shared by all providers."""
from datetime import datetime, timedelta

import pytz
import requests
from bs4 import BeautifulSoup


def get_html_document(url):
    """Get HTML document of given URL."""
    headers = {
        "User-Agent":
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 Safari/537.36", # noqa
        "Accept":
        "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8", # noqa
    }
    resp = requests.get(url, headers=headers)
    return resp.text


def apply_parser(html_doc, selector, parser):
    """Parse items in HTML document."""
    soup = BeautifulSoup(html_doc, "html.parser")
    return parser(soup.select(selector))


def get_ranks(url, selector, parser):
    """Prase ranks from with URL and selectors."""
    return apply_parser(
        get_html_document(url),
        selector,
        parser,
    )


def localize_time(day_time, timezone="UTC"):
    """Return local time, defaulting to UTC.

    >>> dt = datetime(2018, 11, 17)
    >>> local_dt = localize_time(dt)
    >>> local_dt.tzname()
    'UTC'
    >>> local_dt = localize_time(dt, "Asia/Seoul")
    >>> local_dt.tzname()
    'KST'
    """
    if day_time is None:
        day_time = datetime.now()
    local = pytz.timezone(timezone)
    return day_time.astimezone(local)


def append_date_string(url,
                       day_time,
                       date_key=None,
                       date_format="%Y-%m-%d",
                       trailing_slash=False):
    """Add date string to the URL.

    >>> url = "https://mochart.com"
    >>> dt = datetime(2018, 11, 17)
    >>> append_date_string(url, dt)
    'https://mochart.com/2018-11-17'
    """
    dt_str = day_time.strftime(date_format)
    slash = ""
    if trailing_slash:
        slash = "/"
    if date_key:
        return f"{url}?{date_key}={dt_str}{slash}"
    return f"{url}/{dt_str}{slash}"


def get_week_dates(day_time):
    """Return the beginning and end of the week which given time falls to.

    >>> dt = datetime(2018, 11, 17)
    >>> beg, end = get_week_dates(dt)
    >>> beg.strftime("%Y-%m-%d")
    '2018-11-12'
    >>> end.strftime("%Y-%m-%d")
    '2018-11-18'
    """
    beg = day_time - timedelta(days=day_time.weekday())
    end = beg + timedelta(days=6)
    return beg, end


def get_weeks_in(day_time, start_zero=False, start_sunday=False):
    """Return the number of weeks into given month.

    >>> dt = datetime(2018, 1, 1)
    >>> get_weeks_in(dt)
    1
    """
    current_week = int(get_weeks(day_time))
    beginning_week = int(get_weeks(datetime(day_time.year, day_time.month, 1)))
    addition = 0 if start_zero else 1
    return current_week - beginning_week + addition


def get_weeks(day_time, start_sunday=False):
    """Return the number of weeks into given month in string.

    >>> dt = datetime(2018, 1, 1)
    >>> get_weeks(dt)
    '01'
    """
    dt = localize_time(day_time)
    week_format = "%U" if start_sunday else "%W"
    return dt.strftime(week_format)


def get_months(day_time):
    """Return month in string.

    >>> dt = datetime(2018, 1, 1)
    >>> get_months(dt)
    '01'
    """
    dt = localize_time(day_time)
    return dt.strftime("%m")


def get_years(day_time):
    """Return year in string.

    >>> dt = datetime(2018, 1, 1)
    >>> get_years(dt)
    '2018'
    """
    dt = localize_time(day_time)
    return dt.strftime("%Y")


def group_multiples(links):
    """Join links by comma if more than one.

    >>> from collections import namedtuple
    >>> Link = namedtuple("Link", ["text"])
    >>> artists = [Link("IU"), Link("Sungha Jung")]
    >>> group_multiples(artists)
    'IU, Sungha Jung'
    """
    return ", ".join([link.text.strip() for link in links])


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from mochart import utils


BASE_URL = "http://www.gaonchart.co.kr/main/section/chart/online.gaon?nationGbn=T&serviceGbn=ALL"
SELECTOR = "td.subject [title]"
TIMEZONE = "Asia/Seoul"


def parser(rows):
    """Parse texts accordingly from Gaon table."""
    titles = map(lambda t: t[1].text,
                 filter(lambda x: x[0] % 2 == 0, enumerate(rows)))
    artists = map(lambda t: t[1].text.split("|")[0],
                  filter(lambda x: x[0] % 2 != 0, enumerate(rows)))
    albums = map(lambda t: t[1].text.split("|")[1],
                 filter(lambda x: x[0] % 2 != 0, enumerate(rows)))
    return [{
        "title": title,
        "artist": artist,
        "album": album
    } for title, artist, album in zip(titles, artists, albums)]


def week(day_time=None):
    base_url = f"{BASE_URL}&termGbn=week" # noqa
    local_dt = utils.localize_time(day_time, TIMEZONE)
    weeks = utils.get_weeks(local_dt)
    years = utils.get_years(local_dt)
    url = f"{base_url}&targetTime={weeks}&hitYear={years}"
    return utils.get_ranks(url, SELECTOR, parser)


def month(day_time=None):
    base_url = f"{BASE_URL}&termGbn=month" # noqa
    local_dt = utils.localize_time(day_time, TIMEZONE)
    months = utils.get_months(local_dt)
    years = utils.get_years(local_dt)
    url = f"{base_url}&targetTime={months}&hitYear={years}"
    return utils.get_ranks(url, SELECTOR, parser)


def year(day_time=None):
    base_url = f"{BASE_URL}&termGbn=year" # noqa
    local_dt = utils.localize_time(day_time, TIMEZONE)
    years = utils.get_years(local-dt)
    url = f"{base_url}&targetTime={years}&hitYear={years}"
    return utils.get_ranks(url, SELECTOR, parser)



from setuptools import setup, find_packages


setup(
    name="mochart",
    version="0.1.0",
    packages=find_packages(),
    install_requires=[
        "docutils>=0.3",
        "beautifulsoup4==4.6.3",
        "pytest==3.9.3",
        "pytz==2018.7",
        "requests==2.20.0",
    ],

    package_data={
        '': ['*.txt', '*.rst'],
    },

    author="Hyunchel Kim",
    author_email="hyunchel.inbox@gmail.com",
    description="Music online charts.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    license="MIT",
    keywords="music online chart ranking melon mnet naver gaon oricon",
    url="https://hyunchel.github.io/mochart/",
    project_urls={
        "Bug Tracker": "https://github.com/hyunchel/mochart/issues",
        "Documentation": "https://github.com/hyunchel/mochart/README.md",
        "Source Code": "https://github.com/hyunchel/mochart",
    }
)
